@@ -8,8 +8,8 @@ use crate::permissions::Permissions;
 use deno_core::error::bad_resource_id;
 use deno_core::error::type_error;
 use deno_core::error::AnyError;
-use deno_core::op_async;
-use deno_core::op_sync;
+use deno_core::op;
+
 use deno_core::AsyncMutFuture;
 use deno_core::AsyncRefCell;
 use deno_core::Extension;
@@ -29,11 +29,7 @@ use std::os::unix::process::ExitStatusExt;
 
 pub fn init() -> Extension {
   Extension::builder()
-    .ops(vec![
-      ("op_run", op_sync(op_run)),
-      ("op_run_status", op_async(op_run_status)),
-      ("op_kill", op_sync(op_kill)),
-    ])
+    .ops(vec![op_run::decl(), op_run_status::decl(), op_kill::decl()])
     .build()
 }
 
@@ -102,11 +98,8 @@ struct RunInfo {
   stderr_rid: Option<ResourceId>,
 }
 
-fn op_run(
-  state: &mut OpState,
-  run_args: RunArgs,
-  _: (),
-) -> Result<RunInfo, AnyError> {
+#[op]
+fn op_run(state: &mut OpState, run_args: RunArgs) -> Result<RunInfo, AnyError> {
   let args = run_args.cmd;
   state.borrow_mut::<Permissions>().run.check(&args[0])?;
   let env = run_args.env;
@@ -226,10 +219,10 @@ struct ProcessStatus {
   exit_signal: i32,
 }
 
+#[op]
 async fn op_run_status(
   state: Rc<RefCell<OpState>>,
   rid: ResourceId,
-  _: (),
 ) -> Result<ProcessStatus, AnyError> {
   let resource = state
     .borrow_mut()
@@ -303,6 +296,7 @@ pub fn kill(pid: i32, signal: &str) -> Result<(), AnyError> {
   }
 }
 
+#[op]
 fn op_kill(
   state: &mut OpState,
   pid: i32,
